"""
Tests encrypted connection to Alphatronics UNii.
"""

import asyncio
import json
import logging
import unittest

from unii import DEFAULT_PORT, UNiiLocal

from . import async_test

_SETTINGS_JSON = "settings.json"

_LOGGER = logging.getLogger(__name__)


class Test(unittest.TestCase):
    """
    Unit test for encrypted connecting to Alphatronics UNii.
    """

    _host = None
    _port: int = DEFAULT_PORT

    def setUp(self):
        # logging.basicConfig(
        #     format="%(asctime)s %(levelname)-8s %(message)s", level=logging.DEBUG
        # )
        with open(_SETTINGS_JSON, encoding="utf8") as settings_file:
            settings = json.load(settings_file)
            self._host = settings.get("host")
            self._port = settings.get("encrypted_port", self._port)

    @async_test
    async def test_connect_byte_array(self):
        """
        Test connecting to Alphatronics UNii.
        """
        shared_key = b"halloditiseenkey"

        await asyncio.sleep(1)
        unii = UNiiLocal(self._host, self._port, shared_key)
        try:
            result = await unii.test_connection()
            self.assertTrue(result, "Failed to connect to UNii")
        finally:
            # await asyncio.sleep(1)
            await unii.disconnect()

    @async_test
    async def test_connect_string_encode(self):
        """
        Test connecting to Alphatronics UNii.
        """
        shared_key = "halloditiseenkey".encode()

        await asyncio.sleep(1)
        unii = UNiiLocal(self._host, self._port, shared_key)
        try:
            result = await unii.test_connection()
            self.assertTrue(result, "Failed to connect to UNii")
        finally:
            # await asyncio.sleep(1)
            await unii.disconnect()

    @async_test
    async def test_connect_hex_string(self):
        """
        Test connecting to Alphatronics UNii.
        """
        shared_key = "68616C6C6F646974697365656E6B6579"

        await asyncio.sleep(1)
        unii = UNiiLocal(self._host, self._port, shared_key)
        try:
            result = await unii.test_connection()
            self.assertTrue(result, "Failed to connect to UNii")
        finally:
            # await asyncio.sleep(1)
            await unii.disconnect()

    @async_test
    async def test_connect_bytes_fromhex(self):
        """
        Test connecting to Alphatronics UNii.
        """
        shared_key = bytes.fromhex("68616C6C6F646974697365656E6B6579")

        await asyncio.sleep(1)
        unii = UNiiLocal(self._host, self._port, shared_key)
        try:
            result = await unii.test_connection()
            self.assertTrue(result, "Failed to connect to UNii")
        finally:
            # await asyncio.sleep(1)
            await unii.disconnect()


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# pylint: disable=protected-access
# pylint: disable=R0801
"""
Tests automatic reconnect to Alphatronics UNii.
"""

import asyncio
import json
import logging
import unittest

from unii import UNiiLocal

from . import async_test

_SETTINGS_JSON = "settings.json"

_LOGGER = logging.getLogger(__name__)


class Test(unittest.TestCase):
    """
    Unit test for automatic reconnect to Alphatronics UNii.
    """

    _host = None
    _port: int = 6502
    _shared_key = None

    def setUp(self):
        # logging.basicConfig(
        #     format="%(asctime)s %(levelname)-8s %(message)s", level=logging.DEBUG
        # )
        with open(_SETTINGS_JSON, encoding="utf8") as settings_file:
            settings = json.load(settings_file)
            self._host = settings.get("host")
            self._port = settings.get("encrypted_port", self._port)
            self._shared_key = settings.get("shared_key")

    @async_test
    async def test_poll_alive_timeout(self):
        """
        Test Poll Alive timeout.

        To keep the connection alive (and NAT entries active) a poll message has to be sent every
        30 seconds if no other messages where sent during the last 30 seconds.

        The UNii takes care of sending this poll message and should keep the connection
        alive.
        """
        await asyncio.sleep(1)
        unii = UNiiLocal(self._host, self._port, self._shared_key)
        try:
            await unii.connect()
            self.assertTrue(unii.connection.is_open, "Connection is not open")
            while True:
                await asyncio.sleep(10)
                with self.subTest(msg="Connection open"):
                    self.assertTrue(unii.connection.is_open, "Connection is not open")
        finally:
            await unii.disconnect()


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
Data classes used by the UNii library.
"""

import logging
import string
from abc import ABC, abstractmethod
from datetime import datetime
from enum import IntEnum, IntFlag, auto
from typing import Final

import semver

from .sia_code import SIACode

logger = logging.getLogger(__name__)

# Helper functions


def bit_position_to_numeric(data: bytes) -> list[int]:
    """
    Returns which bits in an array of bytes are set, LSB first.

    00001010 => [2, 4]
    """
    data = int.from_bytes(data)
    numerics = []
    for i in range(0, 31):
        bit_position: int = pow(2, i)
        if data & bit_position:
            numerics.append(i + 1)
    return numerics


def bcd_encode(data: str) -> bytes:
    """ "
    Encodes numeric string values in BCD format padded with zeroes to the right.

    https://en.wikipedia.org/wiki/Binary-coded_decimal

    Example implementation from:
    """
    assert data.isdigit()

    # Add trailing zeroes
    while len(data) < 16:
        data += "0"
    logger.debug("Data: %s", data)

    return bytes.fromhex(data)


def decode_and_strip(data: bytes) -> str:
    """Decode a byte array to UTF-8 and strip all leading and trailing whitespace and \x00."""
    return data.decode("utf-8", "replace").strip(string.whitespace + "\x00")


def translate_input_number(input_number: int) -> int:
    """Translate input number according to Appendix 4 of the UNii API."""
    if input_number <= 511:
        return input_number + 1
    if 512 <= input_number <= 543:
        return input_number + 189
    # if 544 <= input_number <= 575:
    #     return -1
    if 576 <= input_number <= 639:
        return input_number + 25
    if 640 <= input_number <= 688:
        return input_number + 161
    # if 689 <= input_number <= 705:
    #     return -1
    if 706 <= input_number <= 962:
        return input_number + 295

    return -1


# Generic command data classes


class UNiiData(ABC):
    # pylint: disable=too-few-public-methods
    """
    UNii Base data class.

    All data classes which are used to send or receive data should inherit from this class.
    """


class UNiiSendData(ABC):
    # pylint: disable=too-few-public-methods
    """
    Method which should be implemented by data classes which are used to send data.
    """

    @abstractmethod
    def to_bytes(self):
        """
        Converts a message to bytes which can be send to the UNii.
        """
        raise NotImplementedError


class UNiiRawData(UNiiData, UNiiSendData):
    # pylint: disable=too-few-public-methods
    """
    UNii Raw data class.

    This dataclass represents the send or received data as a raw array of binary data and can be
    used when no other data classes are available.
    """

    def __init__(self, raw_data):
        self.raw_data = raw_data

    def to_bytes(self):
        return self.raw_data

    def __repr__(self) -> str:
        return "0x" + self.raw_data.hex()


class UNiiResultCode(IntEnum):
    """
    UNii Result Code data class.

    This data class is used as an ACK or NACK response from both sides for the defined commands.
    """

    # pylint: disable=too-few-public-methods

    OK: Final = 0x0000
    ERROR: Final = 0x0001

    def __init__(self, data: bytes):
        if isinstance(data, bytes):
            data = int.from_bytes(data)
        super().__init__(data)


# Equipment related
class UNiiEquipmentInformation(UNiiData):
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    """
    UNii Equipment Information data class.

    This data class contains the response of the "Request Equipment Information" command.
    """

    def __init__(self, data: bytes):
        # pylint: disable=consider-using-min-builtin
        # pylint: disable=too-many-locals
        """ """
        # Version
        version = data[1]
        if version not in [2, 3]:
            raise ValueError("Invalid message version")

        if version == 2:
            software_version = decode_and_strip(data[2:7])
            # The software version can be truncated, to make it a valid SemVer add 0 if the
            # software version ends with .
            if software_version.endswith("."):
                software_version += "0"

            software_date = decode_and_strip(data[7:19])
            self.software_date = datetime.strptime(software_date, "%d-%m-%Y").date()
        elif version == 3:
            software_version = decode_and_strip(data[2:19])
            self.software_date = None

        try:
            self.software_version = semver.Version.parse(software_version)
        except ValueError as ex:
            # Fall back to version 0.0.0 when SemVer can not be parsed.
            self.software_version = semver.Version(0, 0, 0)
            logger.warning(ex)

        device_name_length = data[19]
        self.device_name = decode_and_strip(data[20 : 20 + device_name_length])
        data = data[20 + device_name_length :]

        self.max_inputs = int.from_bytes(data[0:2])
        self.max_groups = data[2]
        self.max_sections = data[3]
        self.max_users = int.from_bytes(data[4:6])

        if version == 3:
            device_id_length = data[6]
            self.device_id = decode_and_strip(data[7 : 7 + device_id_length])
            self.serial_number = self.device_id[:9]
            mac_address = self.device_id[-12:]
            self.mac_address = ":".join(
                mac_address.lower()[i : i + 2] for i in range(0, 12, 2)
            )
        else:
            self.device_id = None
            self.serial_number = None
            self.mac_address = None

    def __str__(self) -> str:
        return str(
            {
                "software_version": self.software_version,
                "software_date": str(self.software_date),
                "device_name": self.device_name,
                "device_id": self.device_id,
                "serial_number": self.serial_number,
                "mac_address": self.mac_address,
            }
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, UNiiEquipmentInformation):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return (
            self.software_version == other.software_version
            and self.software_date == other.software_date
            and self.device_name == other.device_name
            and self.max_inputs == other.max_inputs
            and self.max_groups == other.max_groups
            and self.max_sections == other.max_sections
            and self.max_users == other.max_users
            and self.device_id == other.device_id
            and self.serial_number == other.serial_number
            and self.mac_address == other.mac_address
        )


# Section related
class UNiiSection(dict):
    """
    UNii Section.
    """

    # Get dictionarry keys as attributes.
    __getattr__ = dict.get

    def __init__(self, data: bytes):
        # Version
        version = data[0]
        if version not in [0, 1]:
            raise ValueError("Invalid message version")

        self["active"] = data[1] == 1
        if version == 0:
            name = decode_and_strip(data[2:19])
        elif version == 1:
            name_length = data[2]
            name = decode_and_strip(data[3 : 3 + name_length])
        self["name"] = name


class UNiiSectionArrangement(dict, UNiiData):
    """
    UNii Section Arrangement data class.

    This data class contains the response of the "Request Section Arrangement" command.
    """

    def __init__(self, data: bytes):
        offset = 0
        index = 1
        while offset < len(data):
            # Version
            version = data[0 + offset]
            if version not in [0, 1]:
                raise ValueError("Invalid message version")

            section_length = 19
            if version == 1:
                section_length = data[offset + 2] + 3
            section = UNiiSection(data[offset : offset + section_length])
            section["number"] = index
            self[index] = section

            index += 1
            offset += section_length


class UNiiSectionArmedState(IntEnum):
    """
    The available armed states.
    """

    NOT_PROGRAMMED: Final = 0
    ARMED: Final = 1
    DISARMED: Final = 2
    ALARM: Final = 7
    EXIT_TIMER: Final = 8
    ENTRY_TIMER: Final = 9

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name.lower()


class UNiiSectionStatusRecord(dict):
    """
    UNii Section Status record.
    """

    # Get dictionarry keys as attributes.
    __getattr__ = dict.get

    def __init__(self, data: bytes):
        self["number"] = data[0]
        self["armed_state"] = UNiiSectionArmedState(data[1])


class UNiiSectionStatus(dict, UNiiData):
    """
    UNii Section Status data class.

    This data class contains the response of the "Request Section Status" command.
    """

    # Get dictionarry keys as attributes.
    __getattr__ = dict.get

    def __init__(self, data: bytes):
        # Split data in chunks of 2 bytes
        chunks = [data[pos : pos + 2] for pos in range(0, len(data), 2)]
        # Convert chunks to Section Status Records
        for chunk in chunks:
            section_status_record = UNiiSectionStatusRecord(chunk)
            self[section_status_record.number] = section_status_record


class UNiiArmDisarmSection(UNiiData, UNiiSendData):
    # pylint: disable=too-few-public-methods
    """
    This data class contains the request for "Request Arm Section" and "Request Disarm Section"
    """

    def __init__(self, code: str, number: int):
        self.code = code
        self.number = number

    def to_bytes(self):
        bytes_ = bytearray()
        bytes_.append(0x00)
        bytes_.extend(bcd_encode(self.code))
        bytes_.extend(self.number.to_bytes(1))
        bytes_.append(0x01)
        return bytes(bytes_)


class UNiiReadyToArmState(IntEnum):
    # pylint: disable=too-few-public-methods
    """
    The available Ready To Arm states.
    """

    NOT_PROGRAMMED: Final = 0
    SECTION_ARMED: Final = 1
    # SECTION_DISARMED: Final = 2
    SECTION_READY_FOR_ARMING: Final = 3
    SECTION_NOT_READY_FOR_ARMING: Final = 4
    NOT_AUTHORIZED_TO_ARM: Final = 5
    SYSTEM_ERROR: Final = 6

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name.lower()


class UNiiReadyToArmSectionStatus(UNiiData):
    # pylint: disable=too-few-public-methods
    """
    UNii Ready To Arm Section State data class
    """

    def __init__(self, data: bytes):
        self.section_number = data[0]
        self.section_state = data[1]


class UNiiArmState(IntEnum):
    # pylint: disable=too-few-public-methods
    """
    The available Arm states.
    """

    NO_CHANGE: Final = 0
    SECTION_ARMED: Final = 1
    ARMING_FAILED_SECTION_NOT_READY: Final = 2
    ARMING_FAILED_NOT_AUTHORIZED: Final = 3

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name.lower()


class UNiiArmSectionStatus(UNiiData):
    # pylint: disable=too-few-public-methods
    """
    UNii Arm Section State data class
    """

    def __init__(self, data: bytes):
        self.number = data[0]
        self.arm_state = UNiiArmState(data[1])

    def __repr__(self) -> str:
        return str({"number": self.number, "arm_state": self.arm_state})


class UNiiDisarmState(IntEnum):
    # pylint: disable=too-few-public-methods
    """
    The available Disarm states.
    """

    NO_CHANGE: Final = 0
    SECTION_DISARMED: Final = 1
    DISARMING_FAILED: Final = 2
    DISARMING_FAILED_NOT_AUTHORIZED: Final = 3

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name.lower()


class UNiiDisarmSectionStatus(UNiiData):
    # pylint: disable=too-few-public-methods
    """
    UNii Disarm Section State data class
    """

    def __init__(self, data: bytes):
        self.number = data[0]
        self.disarm_state = UNiiDisarmState(data[1])

    def __repr__(self) -> str:
        return str({"number": self.number, "disarm_state": self.disarm_state})


# Input related


class UNiiInputType(IntEnum):
    """
    The available input types.
    """

    WIRED: Final = auto()
    KEYPAD: Final = auto()
    SPARE: Final = auto()
    WIRELESS: Final = auto()
    KNX: Final = auto()
    DOOR: Final = auto()

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name.lower()


class UNiiSensorType(IntEnum):
    """
    The different kind of sensor types.
    """

    NOT_ACTIVE: Final = 0
    BURGLARY: Final = 1
    FIRE: Final = 2
    TAMPER: Final = 3
    HOLDUP: Final = 4
    MEDICAL: Final = 5
    GAS: Final = 6
    WATER: Final = 7
    TECHNICAL: Final = 8
    DIRECT_DIALER_INPUT: Final = 9
    KEYSWITCH: Final = 10
    NO_ALARM: Final = 11
    EN54_FIRE: Final = 12
    EN54_FIRE_MCP: Final = 13
    EN54_FAULT: Final = 14
    GLASSBREAK: Final = 15

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name.lower()


class UNiiReaction(IntEnum):
    """
    The different kind of reactions.
    """

    DIRECT: Final = 0
    DELAYED: Final = 1
    FOLLOWER: Final = 2
    TWENT_FOUR_HOUR: Final = 3
    LAST_DOOR: Final = 4
    DELAYED_ALARM: Final = 5

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name.lower()


class UNiiInput(dict):
    """
    UNii Input.
    """

    # Get dictionarry keys as attributes.
    __getattr__ = dict.get

    def __init__(self, data: bytes):
        input_number = int.from_bytes(data[0:2])
        self["number"] = input_number

        input_type = UNiiInputType.SPARE
        if 1 <= input_number <= 512:
            input_type = UNiiInputType.WIRED
        elif 701 <= input_number <= 732:
            input_type = UNiiInputType.KEYPAD
        elif 601 <= input_number <= 664:
            input_type = UNiiInputType.WIRELESS
        elif 801 <= input_number <= 845:
            input_type = UNiiInputType.KNX
        elif 1001 <= input_number <= 1128:
            input_type = UNiiInputType.DOOR
        self["input_type"] = input_type

        self["sensor_type"] = UNiiSensorType(data[2])
        self["reaction"] = UNiiReaction(data[3])
        name_length = data[4]
        name = None
        if name_length > 0:
            name = decode_and_strip(data[5 : 5 + name_length])
        self["name"] = name
        self["sections"] = bit_position_to_numeric(data[5 + name_length :])
        self["status"] = UNiiInputState.DISABLED


class UNiiInputArrangement(dict, UNiiData):
    # pylint: disable=too-few-public-methods
    """
    UNii Input Arrangement data class.

    This data class contains the response of the "Request Input Arrangement" command.
    """

    def __init__(self, data: bytes):
        """ """
        # Version
        version = data[1]
        if version != 2:
            raise ValueError("Invalid message version")

        # Block Number
        block_number = int.from_bytes(data[2:4])

        if block_number == 0xFFFF:
            raise ValueError("Invalid block number")

        self.block_number = block_number

        offset = 4
        while offset < len(data):
            name_length = data[4 + offset]
            input_information = data[offset : 9 + offset + name_length]
            input_information = UNiiInput(input_information)
            self[input_information.number] = input_information

            offset += 9 + name_length

    def __str__(self) -> str:
        return str({"block_number": self.block_number, "inputs": super().__str__()})


class UNiiInputState(IntEnum):
    """
    The available input states.
    """

    INPUT_OK: Final = 0x0
    ALARM: Final = 0x1
    TAMPER: Final = 0x2
    MASKING: Final = 0x4
    DISABLED: Final = 0xF

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name.lower()


class UNiiInputStatusRecord(dict):
    """
    UNii Input Status record.
    """

    # Get dictionarry keys as attributes.
    __getattr__ = dict.get

    def __init__(self, data: int):
        self["status"] = UNiiInputState(data & 0x0F)
        self["bypassed"] = data & 0b00010000 == 0b00010000
        self["alarm_memorized"] = data & 0b00100000 == 0b00100000
        self["low_battery"] = data & 0b01000000 == 0b01000000
        self["supervision"] = data & 0b10000000 == 0b10000000


class UNiiInputStatus(dict, UNiiData):
    # pylint: disable=too-few-public-methods
    """
    UNii Input Status data class.
    """

    def __init__(self, data: bytes):
        # Version
        version = data[1]
        if version != 2:
            raise ValueError()

        for index, input_status in enumerate(data[2:]):
            input_status = UNiiInputStatusRecord(input_status)
            input_status["number"] = translate_input_number(index)

            if input_status.number >= 0:
                self[input_status.number] = input_status


class UNiiInputStatusUpdate(UNiiInputStatusRecord, UNiiData):
    # pylint: disable=too-few-public-methods
    """
    UNii Input Status Update data class.
    """

    # Get dictionarry keys as attributes.
    __getattr__ = dict.get

    def __init__(self, data: bytes):
        # Version
        version = data[1]
        if version != 2:
            raise ValueError()

        super().__init__(data[4])
        self["number"] = int.from_bytes(data[2:4])


# Output related


class UNiiOutputType(IntEnum):
    """
    The available output types.
    """

    NOT_ACTIVE: Final = 0
    DIRECT: Final = 1
    TIMED: Final = 2
    FOLLOW_INPUT: Final = 3


class UNiiOutput(dict):
    """
    UNii Output.
    """

    # Get dictionarry keys as attributes.
    __getattr__ = dict.get

    def __init__(self, data: bytes):
        output_number = int.from_bytes(data[0:2])
        self["number"] = output_number

        self["type"] = UNiiOutputType(data[2])

        name_length = data[3]
        name = None
        if name_length > 0:
            name = decode_and_strip(data[4 : 4 + name_length])
        self["name"] = name
        self["sections"] = bit_position_to_numeric(data[5 + name_length :])


class UNiiOutputArrangement(dict, UNiiData):
    # pylint: disable=too-few-public-methods
    """
    UNii Output Arrangement data class.

    This data class contains the response of the "Request Output Arrangement" command.
    """

    def __init__(self, data: bytes):
        """ """
        # Version
        version = data[1]
        if version != 1:
            raise ValueError("Invalid message version")

        # Block Number
        block_number = int.from_bytes(data[2:4])

        if block_number == 0xFFFF:
            raise ValueError("Invalid block number")

        self.block_number = block_number

        offset = 4
        while offset < len(data):
            name_length = data[3 + offset]
            output_information = data[offset : 8 + offset + name_length]
            output_information = UNiiOutput(output_information)
            self[output_information.number] = output_information

            offset += 8 + name_length

    def __str__(self) -> str:
        return str({"block_number": self.block_number, "inputs": super().__str__()})


# Device related


class UNiiDeviceStatusRecord(IntFlag):
    """
    UNii Device Status Record
    """

    LAN_CONNECTION_FAILURE: Final = 32762
    POWER_UNIT_FAILURE_RESTORED: Final = 16384
    POWER_UNIT_FAILURE: Final = 8192
    BATTERY_FAULT_RESTORED: Final = 4096
    BATTERY_FAULT: Final = 2048
    BATTERY_MISSING_RESTORED: Final = 1024
    BATTERY_MISSING: Final = 512
    DEVICE_PRESENT: Final = 256
    RS485_BUS_COMMUNICATION_FAILURE_RESTORED: Final = 128
    RS485_BUS_COMMUNICATION_FAILURE: Final = 64
    TAMPER_SWITCH_OPEN_RESTORED: Final = 32
    TAMPER_SWITCH_OPEN: Final = 16
    LOW_BATTERY_RESTORED: Final = 8
    LOW_BATTERY: Final = 4
    MAINS_FAILURE_RESTORED: Final = 2
    MAINS_FAILURE: Final = 1

    def __repr__(self) -> str:
        return str(self.name)


class UNiiDeviceStatus(UNiiData):
    # pylint: disable=too-few-public-methods
    """
    UNii Device Status data class.

    This data class contains the response of the "Request Device Status" command.
    """

    io_devices: list[UNiiDeviceStatusRecord]
    keyboard_devices: list[UNiiDeviceStatusRecord]
    wiegand_devices: list[UNiiDeviceStatusRecord]
    uwi_devices: list[UNiiDeviceStatusRecord]

    def __init__(self, data: bytes):
        # Version
        version = data[1]
        if version != 2:
            raise ValueError("Invalid message version")

        # Split data in chunks of 2 bytes
        chunks = [data[pos : pos + 2] for pos in range(2, len(data), 2)]
        # Convert chunks to list of Device Status Records
        device_status_records: list[UNiiDeviceStatusRecord] = [
            UNiiDeviceStatusRecord.from_bytes(chunk) for chunk in chunks
        ]

        # Control Panel
        self.control_panel = device_status_records[0]

        # IO Devices
        self.io_devices = device_status_records[1:16]

        # Keyboard Devices
        self.keyboard_devices = device_status_records[16:32]

        # Wiegand Devices
        self.wiegand_devices = device_status_records[32:48]

        # KNX Device
        self.knx_device = device_status_records[48]

        # UWI Devices
        self.uwi_devices = device_status_records[49:51]

        # Redundant Device
        redundant_device = None
        if len(device_status_records) == 52:
            redundant_device = device_status_records[51]
        self.redundant_device = redundant_device

    def __str__(self) -> str:
        return str(
            {
                "control_panel": self.control_panel,
                "io_devices": self.io_devices,
                "keyboard_devices": self.keyboard_devices,
                "wiegand_devices": self.wiegand_devices,
                "knx_device": self.knx_device,
                "uwi_devices": self.uwi_devices,
                "redundant_device": self.redundant_device,
            }
        )


# Event related


class UNiiEventRecord(UNiiData):
    """
    UNii Event Record data class.
    """

    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes

    event_description: str | None = None
    user_number: int | None = None
    user_name: str | None = None
    input_number: int | None = None
    input_name: str | None = None
    device_number: int | None = None
    device_name: str | None = None
    bus_number: int | None = None
    sections: list[int] | None = None
    sia_code: SIACode | None = None

    def __init__(self, data: bytes):
        # pylint: disable=consider-using-min-builtin
        # pylint: disable=too-many-locals
        """ """
        # Version
        version = data[1]
        if version != 3:
            raise ValueError("Invalid message version")

        # Event Number
        self.event_number = int.from_bytes(data[2:4])

        # Timestamp
        year = 1900 + data[4]
        month = data[5] + 1
        day = data[6]
        hour = data[7]
        minute = data[8]
        second = data[9]
        self.timestamp = datetime(year, month, day, hour, minute, second)

        data = data[10:]

        # Description
        event_description_length = data[0]
        if event_description_length > 0:
            self.event_description = decode_and_strip(
                data[1 : 1 + event_description_length]
            )

        data = data[1 + event_description_length :]

        # User
        user_number = int.from_bytes(data[0:2])
        if user_number > 0:
            self.user_number = user_number

        user_name_length = data[2]
        if user_name_length > 0:
            self.user_name = decode_and_strip(data[3 : 3 + user_name_length])

        data = data[3 + user_name_length :]

        # Input
        input_number = int.from_bytes(data[0:2])
        if input_number > 0:
            self.input_number = input_number

        input_name_length = data[2]
        if input_name_length > 0:
            self.input_name = decode_and_strip(data[3 : 3 + input_name_length])

        data = data[3 + input_name_length :]

        # Device
        device_number = int.from_bytes(data[0:2])
        if device_number > 0:
            self.device_number = device_number

        device_name_length = data[2]
        if device_name_length > 0:
            self.device_name = decode_and_strip(data[3 : 3 + device_name_length])

        data = data[3 + device_name_length :]

        # Bus
        self.bus_number = data[0]

        # Sections
        self.sections = bit_position_to_numeric(data[1:5])

        # SIA Code
        sia_code = decode_and_strip(data[5:7])
        if sia_code != "":
            try:
                self.sia_code = SIACode(sia_code)
            except ValueError as ex:
                logger.warning(ex)

    def __repr__(self) -> str:
        return str(
            {
                "event_number": self.event_number,
                "timestamp": str(self.timestamp),
                "event_description": self.event_description,
                "user_number": self.user_number,
                "user_name": self.user_name,
                "input_number": self.input_number,
                "input_name": self.input_name,
                "device_number": self.device_number,
                "device_name": self.device_name,
                "bus_number": self.bus_number,
                "sections": self.sections,
                "sia_code": self.sia_code,
            }
        )



"""
Classes for interfacing with Alphatronics UNii security systems.
"""

import asyncio
import logging
import time
from abc import ABC
from datetime import datetime, timedelta
from enum import IntFlag, auto
from threading import Lock
from typing import Any, Final

from .task_helper import save_task_reference
from .unii_command import UNiiCommand
from .unii_command_data import (
    UNiiArmDisarmSection,
    UNiiArmSectionStatus,
    UNiiArmState,
    UNiiData,
    UNiiDeviceStatus,
    UNiiDisarmSectionStatus,
    UNiiDisarmState,
    UNiiEquipmentInformation,
    UNiiInput,
    UNiiInputArrangement,
    UNiiInputState,
    UNiiInputStatus,
    UNiiInputStatusRecord,
    UNiiOutput,
    UNiiRawData,
    UNiiSection,
    UNiiSectionArmedState,
    UNiiSectionArrangement,
    UNiiSectionStatus,
)
from .unii_connection import (
    DEFAULT_PORT,
    UNiiConnection,
    UNiiConnectionError,
    UNiiTCPConnection,
)
from .unii_message import UNiiResponseMessage

logger = logging.getLogger(__name__)

_POLL_ALIVE_INTERVAL: Final = timedelta(seconds=30)


class UNiiEncryptionError(Exception):
    """"""


class UNiiFeature(IntFlag):
    """Features implemented by the UNii library."""

    ARM_SECTION: Final = auto()
    BYPASS_ZONE: Final = auto()
    SET_OUTPUT: Final = auto()


class UNii(ABC):
    """
    UNii base class for interfacing with Alphatronics UNii security systems.
    """

    unique_id: str | None = None
    model: str = "Unknown"
    connected: bool = False

    equipment_information: UNiiEquipmentInformation | None = None
    sections: dict[int, UNiiSection]
    inputs: dict[int, UNiiInput]
    outputs: dict[int, UNiiOutput]
    device_status: UNiiDeviceStatus | None = None

    connection: UNiiConnection

    features: list[UNiiFeature]

    _event_occurred_callbacks: list[Any]

    def __init__(
        self,
    ):
        super().__init__()

        self.sections = {}
        self.inputs = {}
        self.outputs = {}

        self.features = []

        self._event_occurred_callbacks = []

    async def test_connection(self) -> bool:
        """
        Connect to Alphatronics UNii

        Throws UNiiConnectionError when no connection can be established.
        Returns False when no connection can be established
        """
        raise NotImplementedError

    async def connect(self) -> bool:
        """
        Connect to Alphatronics UNii

        Throws UNiiEncryptionError when no encrypted message can be send.
        Returns False when no connection can be established
        """
        raise NotImplementedError

    async def disconnect(self) -> bool:
        """
        Disconnect from Alphatronics UNii
        """
        raise NotImplementedError

    def add_event_occurred_callback(self, callback):
        """
        Adds an Event Occurred Callback to the UNii.
        """
        self._event_occurred_callbacks.append(callback)

    def _forward_to_event_occurred_callbacks(
        self, command: UNiiCommand, data: UNiiData | None
    ):
        for callback in self._event_occurred_callbacks:
            try:
                callback(command, data)
            # pylint: disable=broad-exception-caught
            except Exception as ex:
                logger.error("Exception in Event Occurred Callback: %s", ex)

    async def arm_section(self, number: int, code: str) -> bool:
        """Arm a section."""
        raise NotImplementedError

    async def disarm_section(self, number: int, code: str) -> bool:
        """Disarm a section."""
        raise NotImplementedError


class UNiiLocal(UNii):
    # pylint: disable=too-many-instance-attributes
    """
    UNii class for interfacing with Alphatronics UNii security systems on the local
    network.
    """

    _received_message_queue: dict[int, list[Any]]
    _waiting_for_message: dict[int, UNiiCommand | None]

    _poll_alive_task: asyncio.Task | None = None

    def __init__(
        self, host: str, port: int = DEFAULT_PORT, shared_key: bytes | None = None
    ):
        super().__init__()

        # If the shared key is provided as hex string convert it to bytes.
        if shared_key is not None and isinstance(shared_key, str):
            shared_key = bytes.fromhex(shared_key)
        self.connection = UNiiTCPConnection(host, port, shared_key)
        self.unique_id = f"{host}:{port}"

        self._received_message_queue = {}
        self._waiting_for_message = {}

        self._received_message_queue_lock = Lock()

    async def test_connection(self) -> bool:
        success = False

        await self.connection.connect()

        self.connection.set_message_received_callback(self._message_received_callback)
        response, _ = await self._send_receive(
            UNiiCommand.CONNECTION_REQUEST,
            None,
            UNiiCommand.CONNECTION_REQUEST_RESPONSE,
            False,
        )

        if response is None and self.connection.is_encrypted:
            await self.disconnect()
            raise UNiiEncryptionError()

        if response == UNiiCommand.CONNECTION_REQUEST_RESPONSE:
            response, data = await self._send_receive(
                UNiiCommand.REQUEST_EQUIPMENT_INFORMATION,
                None,
                UNiiCommand.RESPONSE_REQUEST_EQUIPMENT_INFORMATION,
                False,
            )
            if (
                response is None
                or data is None
                or response != UNiiCommand.RESPONSE_REQUEST_EQUIPMENT_INFORMATION
                or self.equipment_information is None
            ):
                logger.error("Failed to retrieve equipment information.")
            else:
                success = True

        await self.disconnect()
        return success

    async def _connect(self) -> bool:
        await self.connection.connect()

        self.connection.set_message_received_callback(self._message_received_callback)
        response, _ = await self._send_receive(
            UNiiCommand.CONNECTION_REQUEST,
            None,
            UNiiCommand.CONNECTION_REQUEST_RESPONSE,
            False,
        )

        if response is None and self.connection.is_encrypted:
            await self.disconnect()
            raise UNiiEncryptionError()

        if response == UNiiCommand.CONNECTION_REQUEST_RESPONSE:
            response, data = await self._send_receive(
                UNiiCommand.REQUEST_EQUIPMENT_INFORMATION,
                None,
                UNiiCommand.RESPONSE_REQUEST_EQUIPMENT_INFORMATION,
                False,
            )
            if (
                response is None
                or data is None
                or response != UNiiCommand.RESPONSE_REQUEST_EQUIPMENT_INFORMATION
                or self.equipment_information is None
            ):
                logger.error("Failed to retrieve equipment information.")
                await self._disconnect()
                return False

            await self._send_receive(
                UNiiCommand.REQUEST_SECTION_ARRANGEMENT,
                None,
                UNiiCommand.RESPONSE_REQUEST_SECTION_ARRANGEMENT,
                False,
            )

            await self._send_receive(
                UNiiCommand.REQUEST_SECTION_STATUS,
                UNiiRawData(
                    bytes.fromhex(
                        "0102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f20"
                    )
                ),
                UNiiCommand.RESPONSE_REQUEST_SECTION_STATUS,
                False,
            )

            block = 0
            while True:
                block += 1
                _, data = await self._send_receive(
                    UNiiCommand.REQUEST_INPUT_ARRANGEMENT,
                    UNiiRawData(block.to_bytes(2)),
                    UNiiCommand.RESPONSE_REQUEST_INPUT_ARRANGEMENT,
                    False,
                )
                if data is None:
                    break

            await self._send_receive(
                UNiiCommand.REQUEST_INPUT_STATUS,
                None,
                UNiiCommand.INPUT_STATUS_CHANGED,
                False,
            )

            software_version = (
                self.equipment_information.software_version.finalize_version()
            )
            if software_version.match(">=2.17.0"):
                block = 0
                while True:
                    block += 1
                    _, data = await self._send_receive(
                        UNiiCommand.REQUEST_OUTPUT_ARRANGEMENT,
                        UNiiRawData(block.to_bytes(2)),
                        UNiiCommand.RESPONSE_REQUEST_OUTPUT_ARRANGEMENT,
                        False,
                    )
                    if data is None:
                        break

            await self._send_receive(
                UNiiCommand.REQUEST_DEVICE_STATUS,
                None,
                UNiiCommand.DEVICE_STATUS_CHANGED,
                False,
            )

            self.connected = True

            self._forward_to_event_occurred_callbacks(
                UNiiCommand.CONNECTION_REQUEST_RESPONSE, None
            )

            return True
        return False

    async def connect(self) -> bool:
        try:
            if await self._connect():
                self._poll_alive_task = asyncio.create_task(
                    self._poll_alive_coroutine()
                )
                save_task_reference(self._poll_alive_task)
                return True
        except UNiiConnectionError as ex:
            logger.error(str(ex))

        return False

    async def _disconnect(self) -> bool:
        try:
            await self._send(UNiiCommand.NORMAL_DISCONNECT, None, False)
        except UNiiEncryptionError:
            pass

        if await self.connection.close():
            self.connected = False
            # Re-using the Normal Disconnect command to let the Event Occurred Callbacks know the
            # UNii is disconnected.
            self._forward_to_event_occurred_callbacks(
                UNiiCommand.NORMAL_DISCONNECT, None
            )
            return True

        return False

    async def disconnect(self) -> bool:
        await self._cancel_poll_alive()

        if self.connection is not None and self.connection.is_open:
            try:
                if not await self._disconnect():
                    return False
            except UNiiConnectionError as ex:
                logger.error(str(ex))
                return False

        if self._poll_alive_task is not None:
            await self._cancel_poll_alive()
        return True

    async def _send(
        self, command: UNiiCommand, data: UNiiData | None = None, reconnect: bool = True
    ) -> int | None:
        try:
            if self.connection is None and reconnect:
                logger.info("Trying to connect")
                await self._connect()
            elif not self.connection.is_open and reconnect:
                logger.info("Trying to reconnect")
                await self._connect()
            elif self.connection is None or not self.connection.is_open:
                # ToDo: Throw exception?
                return None

            # logger.debug("Sending command %s", command)
            return await self.connection.send(command, data)
        except UNiiEncryptionError:
            self._forward_to_event_occurred_callbacks(UNiiCommand.REAUTHENTICATE, None)
            raise

    async def _send_receive(
        self,
        command: UNiiCommand,
        data: UNiiData | None = None,
        expected_response: UNiiCommand | None = None,
        reconnect: bool = True,
    ) -> list[Any]:
        tx_sequence = await self._send(command, data, reconnect)
        if tx_sequence is not None:
            return await self._get_received_message(tx_sequence, expected_response)
        return [None, None]

    def _handle_equipment_information(self, data: UNiiEquipmentInformation):
        if (
            self.equipment_information is not None
            and self.equipment_information != data
        ):
            self._forward_to_event_occurred_callbacks(
                UNiiCommand.RELOAD_CONFIGURATION, None
            )

        self.equipment_information = data

        # Get capabilities based on firmware version number
        # Library is currently read-only, so disabled for now
        # software_version = (
        #     self.equipment_information.software_version.finalize_version()
        # )
        # if software_version.match(">=2.17.0"):
        #     self.features.append(UNiiFeature.ARM_SECTION)
        #     self.features.append(UNiiFeature.BYPASS_ZONE)
        #     self.features.append(UNiiFeature.SET_OUTPUT)

    def _handle_section_arrangement(self, data: UNiiSectionArrangement):
        for _, section in data.items():
            if section.number not in self.sections and section.active:
                self.sections[section.number] = section
            elif section.number in self.sections:
                self.sections[section.number].update(section)

    def _handle_section_status(self, data: UNiiSectionStatus):
        for _, section_status in data.items():
            if section_status.number in self.sections:
                section_status["active"] = (
                    section_status.armed_state != UNiiSectionArmedState.NOT_PROGRAMMED
                )
                self.sections[section_status.number].update(section_status)
            elif section_status.armed_state != UNiiSectionArmedState.NOT_PROGRAMMED:
                # This should never happen
                logger.warning(
                    "Status for unknown section %i changed", section_status.number
                )

    def _handle_input_status_update(self, input_status: UNiiInputStatusRecord):
        if input_status.number in self.inputs:
            self.inputs[input_status.number].update(input_status)
        elif input_status.status != UNiiInputState.DISABLED:
            # This should never happen
            logger.warning("Status for unknown input %i changed", input_status.number)

    def _handle_input_status_changed(self, data: UNiiInputStatus):
        for _, input_status in data.items():
            self._handle_input_status_update(input_status)

    def _handle_input_arrangement(self, data: UNiiInputArrangement):
        if data is None:
            return
        for _, unii_input in data.items():
            # Expand sections
            for index, section in enumerate(unii_input.sections):
                unii_input["sections"][index] = self.sections[section]

            if unii_input.number not in self.inputs:
                self.inputs[unii_input.number] = unii_input
            else:
                # Retain the input status before updating the input with new data.
                unii_input.status = self.inputs[unii_input.number].status
                self.inputs[unii_input.number].update(unii_input)

    async def _message_received_callback(
        self, tx_sequence: int, command: UNiiCommand, data: UNiiData
    ):
        match command:
            case UNiiCommand.EVENT_OCCURRED:
                self._forward_to_event_occurred_callbacks(command, data)
                if self.connected:
                    try:
                        await self._send(
                            UNiiCommand.RESPONSE_EVENT_OCCURRED, None, False
                        )
                    except UNiiEncryptionError:
                        pass
            case UNiiCommand.INPUT_STATUS_CHANGED:
                self._handle_input_status_changed(data)
            case UNiiCommand.INPUT_STATUS_UPDATE:
                self._handle_input_status_update(data)
            case UNiiCommand.DEVICE_STATUS_CHANGED:
                self.device_status = data
            case UNiiCommand.RESPONSE_REQUEST_SECTION_ARRANGEMENT:
                self._handle_section_arrangement(data)
            case UNiiCommand.RESPONSE_REQUEST_SECTION_STATUS:
                self._handle_section_status(data)
            case UNiiCommand.RESPONSE_REQUEST_INPUT_ARRANGEMENT:
                self._handle_input_arrangement(data)
            case UNiiCommand.RESPONSE_REQUEST_EQUIPMENT_INFORMATION:
                self._handle_equipment_information(data)

        if tx_sequence in self._waiting_for_message and self._waiting_for_message[
            tx_sequence
        ] in [None, command]:
            with self._received_message_queue_lock:
                self._received_message_queue[tx_sequence] = [command, data]

        self._forward_to_event_occurred_callbacks(command, data)

    async def _get_received_message(
        self, tx_sequence: int, expected_response: UNiiCommand | None = None
    ) -> list[Any]:
        timeout = time.time() + 5
        self._waiting_for_message[tx_sequence] = expected_response
        while self.connection.is_open and time.time() < timeout:
            with self._received_message_queue_lock:
                if tx_sequence in self._received_message_queue:
                    return self._received_message_queue.pop(tx_sequence)
            # logger.debug("Waiting for message %i to be received", tx_sequence)
            await asyncio.sleep(0.1)

        logger.error("Message %i was not received", tx_sequence)
        del self._waiting_for_message[tx_sequence]
        return [None, None]

    async def _poll_alive(self) -> bool:
        try:
            response, _ = await self._send_receive(
                UNiiCommand.POLL_ALIVE_REQUEST,
                None,
                UNiiCommand.POLL_ALIVE_RESPONSE,
                True,
            )
            # logger.debug("Response received: %s", response)
            if response == UNiiCommand.POLL_ALIVE_RESPONSE:
                # logger.debug("Poll Alive success")
                return True
        except UNiiConnectionError as ex:
            logger.error(str(ex))

        if self.connection.is_open:
            logger.error("Poll Alive failed")
        return False

    async def _cancel_poll_alive(self) -> bool:
        if self._poll_alive_task is not None and not (
            self._poll_alive_task.done() or self._poll_alive_task.cancelled()
        ):
            self._poll_alive_task.cancel()
            try:
                await self._poll_alive_task
            except asyncio.CancelledError:
                logger.debug("Poll alive task was cancelled")
                self._poll_alive_task = None

        if self._poll_alive_task is not None:
            logger.error("Failed to cancel poll alive task")
            logger.debug("Poll alive task: %s", self._poll_alive_task)
            return False

        return True

    async def _poll_alive_coroutine(self):
        """
        To keep the connection alive (and NAT entries active) a poll message has to be sent every
        30 seconds if no other messages where sent during the last 30 seconds.
        """
        while True:
            try:
                if (
                    not self.connection.is_open
                    or datetime.now()
                    > self.connection.last_message_sent + _POLL_ALIVE_INTERVAL
                ) and not await self._poll_alive():
                    if self.connection.is_open:
                        await self._disconnect()
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                logger.debug("Poll alive coroutine was canceled")
                break
            except UNiiEncryptionError:
                break

        self._poll_alive_task = None
        logger.debug("Poll Alive coroutine stopped")

    async def arm_section(self, number: int, code: str) -> bool:
        """Arm a section."""
        response, data = await self._send_receive(
            UNiiCommand.REQUEST_ARM_SECTION,
            UNiiArmDisarmSection(code, number),
            UNiiCommand.RESPONSE_ARM_SECTION,
        )
        if response == UNiiCommand.RESPONSE_ARM_SECTION and data.arm_state in [
            UNiiArmState.SECTION_ARMED,
            UNiiArmState.NO_CHANGE,
        ]:
            return True

        logger.error("Arming failed: %s", data.arm_state)
        return False

    async def disarm_section(self, number: int, code: str) -> bool:
        """Disarm a section."""
        response, data = await self._send_receive(
            UNiiCommand.REQUEST_DISARM_SECTION,
            UNiiArmDisarmSection(code, number),
            UNiiCommand.RESPONSE_DISARM_SECTION,
        )
        if response == UNiiCommand.RESPONSE_DISARM_SECTION and data.disarm_state in [
            UNiiDisarmState.SECTION_DISARMED,
            UNiiDisarmState.NO_CHANGE,
        ]:
            return True

        if data is not None:
            logger.error("Disarming failed: %s", data.disarm_state)
        else:
            logger.error("Disarming failed")
        return False



# pylint: disable=R0801
"""
Test Input Status.
"""
import asyncio
import json
import logging
import unittest

from unii import UNiiLocal

from . import async_test

_SETTINGS_JSON = "settings.json"

_LOGGER = logging.getLogger(__name__)


class Test(unittest.TestCase):
    """
    Test Input Status.
    """

    _unii = None

    def setUp(self):
        with open(_SETTINGS_JSON, encoding="utf8") as settings_file:
            settings = json.load(settings_file)
            host = settings.get("host")
            port = settings.get("encrypted_port", 6502)
            shared_key = settings.get("shared_key", 6502)

            self._unii = UNiiLocal(host, port, shared_key)

    @async_test
    async def test_outputs(self):
        """
        Test connecting to Alphatronics UNii.
        """
        await asyncio.sleep(1)
        try:
            result = await self._unii.connect()
            self.assertTrue(result, "Failed to connect to UNii")
            self.assertIsNotNone(self._unii.outputs, "Output Status not set")
            for _, unii_output in self._unii.outputs.items():
                _LOGGER.info(unii_output)
        finally:
            await asyncio.sleep(1)
            await self._unii.disconnect()


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# pylint: disable=protected-access
"""
Test UNii message CRC-16 checksums.

Calculating the CRC-16 checksum for validation is done using:
https://crccalc.com/
"""

# import logging
import unittest

from unii.unii_message import _UNiiMessage


class Test(unittest.TestCase):
    """
    Test UNii message checksum correctness.
    """

    def test_checksum_1(self):
        """
        Tests the correct working of the checksum algorithm.
        """
        message = _UNiiMessage()

        crc = message._calculate_crc16(bytes.fromhex("0123456789abcdef"))
        self.assertEqual(crc, 0xA955)

    def test_checksum_2(self):
        """
        Tests the correct working of the checksum algorithm.
        """
        message = _UNiiMessage()

        crc = message._calculate_crc16(bytes.fromhex("0123456789"))
        self.assertEqual(crc, 0x6282)

    def test_checksum_3(self):
        """
        Tests the correct working of the checksum algorithm.
        """
        message = _UNiiMessage()

        crc = message._calculate_crc16(
            bytes.fromhex(
                "ffff08be2c53000000000401002000010000000000000000000000000000"
            )
        )
        self.assertEqual(crc, 0x5ED3)

    def test_checksum_4(self):
        """
        Tests the correct working of the checksum algorithm.
        """
        message = _UNiiMessage()

        crc = message._calculate_crc16(
            bytes.fromhex("f109441a389608be2c530402001400020000")
        )
        self.assertEqual(crc, 0x853E)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



import argparse
import asyncio
import logging
import sys

from unii import UNiiLocal

_LOGGER = logging.getLogger(__name__)


async def main(unii):
    try:
        if not await unii.connect():
            _LOGGER.error("Failed to connect to UNii")
            return 1

        while True:
            await asyncio.sleep(1)
    except KeyboardInterrupt:
        # Handle keyboard interrupt
        pass
    finally:
        await unii.disconnect()

    return 0


if __name__ == "__main__":
    # Read command line arguments
    argparser = argparse.ArgumentParser()

    subparsers = argparser.add_subparsers()

    local_parser = subparsers.add_parser("local")
    local_parser.add_argument("host")
    local_parser.add_argument("port", type=int)
    local_parser.add_argument("key", nargs="?")

    argparser.add_argument("--debug", dest="debugLogging", action="store_true")

    args = argparser.parse_args()

    _LOGGER.error(args)

    if args.debugLogging:
        logging.basicConfig(
            format="%(asctime)s %(levelname)-8s %(filename)s:%(lineno)d %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(format="%(message)s", level=logging.INFO)

    if "host" in args:
        unii = UNiiLocal(args.host, args.port, args.key)
    else:
        sys.exit(1)

    try:
        loop = asyncio.new_event_loop()
        sys.exit(loop.run_until_complete(main(unii)))
    finally:
        loop.close()



"""
Messages used by the UNii library.
"""

import logging
from enum import IntEnum
from typing import Final

from Crypto.Cipher import AES

from .unii_command import UNiiCommand
from .unii_command_data import (
    UNiiArmSectionStatus,
    UNiiData,
    UNiiDeviceStatus,
    UNiiDisarmSectionStatus,
    UNiiEquipmentInformation,
    UNiiEventRecord,
    UNiiInputArrangement,
    UNiiInputStatus,
    UNiiInputStatusUpdate,
    UNiiOutputArrangement,
    UNiiRawData,
    UNiiReadyToArmSectionStatus,
    UNiiResultCode,
    UNiiSectionArrangement,
    UNiiSectionStatus,
)

logger = logging.getLogger(__name__)


class UNiiMessageError(Exception):
    """
    UNii Base Message Error
    """


class UNiiChecksumError(UNiiMessageError):
    """
    UNii Checksum Error.

    When a received message fails validation against the checksum this error is thrown.
    """

    def __init__(self, expected_checksum: int, received_checksum: int):
        self._expected_checksum = expected_checksum
        self._received_checksum = received_checksum

    def __str__(self) -> str:
        return (
            f"Invalid checksum. Expected {self._expected_checksum}, "
            + f"received {self._received_checksum}."
        )


class UNiiIncompleteMessageError(UNiiMessageError):
    """
    UNii Checksum Error.

    When the length of a received message does not equal the message length header field this error
    is thrown.
    """

    def __init__(self, expected_lenght: int, received_lenght: int):
        self._expected_lenght = expected_lenght
        self._received_lenght = received_lenght

    def __str__(self) -> str:
        return (
            f"Incomplete message received. Expected {self._expected_lenght} bytes, "
            + f"received {self._received_lenght} bytes."
        )


class UNiiInvallidMessageError(UNiiMessageError):
    """
    UNii Invallid Message Error.

    When a message can not be parsed this error is thrown.
    """


class UNiiProtocolID(IntEnum):
    """
    UNii Protocol IDs

    Only no encryption and basic encryption are supported, advanced encryption is currently not
    implemented.
    """

    NO_ENCRYPTION: Final = 0x04
    BASIC_ENCRYPTION: Final = 0x05
    # Advanced encryption is currently not implemented.
    # ADVANCED_ENCRYPTION: Final = 0x??

    def __str__(self) -> str:
        return {
            self.NO_ENCRYPTION: "No Encryption",
            self.BASIC_ENCRYPTION: "Basic Encryption",
        }[self]


class UNiiPacketType(IntEnum):
    """
    UNii Packet Types
    """

    SESSION_SETUP: Final = 0x01
    NORMAL_CONECTION: Final = 0x02

    def __str__(self) -> str:
        return {
            self.SESSION_SETUP: "Session Setup",
            self.NORMAL_CONECTION: "Normal Connection",
        }[self]


class _UNiiMessage:
    # pylint: disable=too-few-public-methods
    """
    UNii Message class
    """

    _CRC16_INIT: Final = 0x0000
    _CRC16_POLYNOMIAL: Final = 0x1021
    _CRC16_REFIN: Final = False
    _CRC16_REFOUT: Final = False
    _CRC16_XOROUT: Final = 0x0000

    session_id: int | None = None
    tx_sequence: int | None = None
    rx_sequence: int | None = None
    command: UNiiCommand | None = None
    data: UNiiData | None = None

    def _calculate_crc16(self, message: bytes) -> int:
        """
        Calculates the CRC-16 checksum over the given message
        """
        crc = self._CRC16_INIT
        for _, byte in enumerate(message):
            crc ^= byte << 8
            for _ in range(8):
                if (crc & 0x8000) > 0:
                    crc <<= 1
                    crc ^= self._CRC16_POLYNOMIAL
                else:
                    crc <<= 1
        crc &= 0xFFFF
        # logger.debug("CRC-16 for 0x%s: %s", message.hex(), hex(crc))
        return crc

    def __str__(self) -> str:
        return str(
            {
                "session_id": self.session_id,
                "tx_sequence": self.tx_sequence,
                "rx_sequence": self.rx_sequence,
                "command": self.command,
                "data": self.data,
            }
        )


class UNiiRequestMessage(_UNiiMessage):
    """
    UNii Message class for sending

    All numbers are exchanged in big endian order (Most significant byte first)
    """

    _protocol_id: UNiiProtocolID | None = None

    @property
    def packet_type(self) -> UNiiPacketType | None:
        """
        Get packet type based on command
        """
        if self.command is None:
            return None
        if self.command < 0x0008:
            return UNiiPacketType.SESSION_SETUP
        return UNiiPacketType.NORMAL_CONECTION

    def _create_header(self):
        """
        Creates the header, first 14 bytes, of the message.
        """
        header = bytearray()
        # 2 byte Session ID
        header += self.session_id.to_bytes(2)
        # 4 byte Sequence number
        header += self.tx_sequence.to_bytes(4)
        # 4 byte Last received sequence number
        header += self.rx_sequence.to_bytes(4)
        # 1 byte Protocol ID
        header += self._protocol_id.to_bytes(1)
        # 1 byte Packet Type
        if self.command < 0x0008:
            header += UNiiPacketType.SESSION_SETUP.to_bytes(1)
        else:
            header += UNiiPacketType.NORMAL_CONECTION.to_bytes(1)
        # 2 byte Total Packet Length
        # To be replaced with the actual packet length once the total message is constructed.
        header.append(0x00)
        header.append(0x00)

        return header

    def _create_payload(self):
        """
        Creates the payload of the message, existing of command, data length and data.
        """
        payload = bytearray()

        # 2 byte Command
        payload += self.command.to_bytes(2)

        # 2 byte Data Length
        data_length = 0
        if self.data is not None:
            data_length = len(self.data.to_bytes())
        payload += data_length.to_bytes(2)

        # variable length The actual data
        if self.data is not None:
            payload += self.data.to_bytes()

        return payload

    def to_bytes(self, shared_key: bytes | None = None) -> bytes:
        """
        Converts the UNii Message to a sequence of bytes
        """
        if shared_key is None:
            self._protocol_id = UNiiProtocolID.NO_ENCRYPTION
        else:
            self._protocol_id = UNiiProtocolID.BASIC_ENCRYPTION

        header = self._create_header()
        payload = self._create_payload()

        # Add Padding
        packet_length = len(header) + len(payload) + 2  # Add 2 for the Checksum
        n_padding_bytes = 16 - (packet_length % 16)
        for _ in range(n_padding_bytes):
            payload.append(0x00)

        if shared_key is not None:
            # logger.debug("Payload: 0x%s", payload.hex())
            # logger.debug("Shared Key: 0x%s", shared_key.hex())

            # As Initialization Vector for the encryption the first 12 bytes of the header are used.
            initial_value = header[:12]
            # The block counter is reset to 0 for every new message.
            initial_value += b"\00\00\00\00"
            # logger.debug("Initial value: 0x%s", initial_value.hex())

            aes = AES.new(
                shared_key, AES.MODE_CTR, initial_value=initial_value, nonce=b""
            )
            payload = aes.encrypt(payload)
            # logger.debug("Encrypted Payload: 0x%s", payload.hex())

        message = header + payload

        # Insert Total Packet Length
        packet_length = len(message) + 2  # Add 2 for the Checksum
        packet_length = packet_length.to_bytes(2)
        message[12] = packet_length[0]
        message[13] = packet_length[1]

        # 2 byte Checksum
        crc = self._calculate_crc16(bytes(message))
        message += crc.to_bytes(2)

        # logger.debug("Message: %s", message.hex())
        return bytes(message)


class UNiiResponseMessage(_UNiiMessage):
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-locals
    """
    UNii Message class for receiving
    """

    def __init__(self, message: bytes, shared_key: bytes | None = None):
        # pylint: disable=too-many-statements
        """ """
        assert message is not None

        # First 14 bytes Header
        header = message[:14]
        payload = message[14:-2]
        checksum = message[-2:]

        # Length
        packet_length = int.from_bytes(header[12:14])

        if packet_length != len(message):
            raise UNiiIncompleteMessageError(packet_length, len(message))

        # Last 2 bytes Checksum
        received_checksum = int.from_bytes(checksum)

        expected_checksum = self._calculate_crc16(message[0:-2])
        if received_checksum != expected_checksum:
            raise UNiiChecksumError(expected_checksum, received_checksum)

        # Session ID
        self.session_id = int.from_bytes(header[:2])

        # TX Sequence
        self.tx_sequence = int.from_bytes(header[2:6])
        # RX Sequence
        self.rx_sequence = int.from_bytes(header[6:10])
        # Protocol ID
        protocol_id = UNiiProtocolID(header[10])
        # Packet Type
        # packet_type = UNiiPacketType(header[11])
        # Length
        # message_length = int.from_bytes(header[11:13])

        if protocol_id == UNiiProtocolID.BASIC_ENCRYPTION and shared_key is not None:
            # As Initialization Vector for the encryption the first 12 bytes of the header are used.
            initial_value = header[:12]
            # The block counter is reset to 0 for every new message.
            initial_value += b"\00\00\00\00"

            payload = self._decrypt(shared_key, initial_value, payload)

        # Command
        command = int.from_bytes(payload[:2])

        try:
            self.command = UNiiCommand(command)
        except ValueError as ex:
            logger.warning(ex)
            self.command = None

        # Data length
        data_length = int.from_bytes(payload[2:4])

        # Data
        data: bytes | UNiiData | None = None
        if data_length > 0:
            data = payload[4 : 4 + data_length]
            # logger.debug("%s data: %i bytes, 0x%s", self.command, len(data), data.hex())
            try:
                match self.command:
                    # Generic
                    case UNiiCommand.GENERAL_RESPONSE:
                        data = UNiiResultCode(data)
                    # Equipment related
                    case UNiiCommand.RESPONSE_REQUEST_EQUIPMENT_INFORMATION:
                        data = UNiiEquipmentInformation(data)
                    # Section related
                    case UNiiCommand.RESPONSE_REQUEST_SECTION_ARRANGEMENT:
                        data = UNiiSectionArrangement(data)
                    case UNiiCommand.RESPONSE_REQUEST_SECTION_STATUS:
                        data = UNiiSectionStatus(data)
                    case UNiiCommand.RESPONSE_READY_TO_ARM_SECTIONS:
                        data = UNiiReadyToArmSectionStatus(data)
                    case UNiiCommand.RESPONSE_ARM_SECTION:
                        data = UNiiArmSectionStatus(data)
                    case UNiiCommand.RESPONSE_DISARM_SECTION:
                        data = UNiiDisarmSectionStatus(data)
                    # Input related
                    case UNiiCommand.RESPONSE_REQUEST_INPUT_ARRANGEMENT:
                        data = UNiiInputArrangement(data)
                    case UNiiCommand.INPUT_STATUS_CHANGED:
                        data = UNiiInputStatus(data)
                    case UNiiCommand.INPUT_STATUS_UPDATE:
                        data = UNiiInputStatusUpdate(data)
                    # Output related
                    case UNiiCommand.RESPONSE_REQUEST_OUTPUT_ARRANGEMENT:
                        data = UNiiOutputArrangement(data)
                    # Device related
                    case UNiiCommand.DEVICE_STATUS_CHANGED:
                        data = UNiiDeviceStatus(data)
                    # Event related
                    case UNiiCommand.EVENT_OCCURRED:
                        data = UNiiEventRecord(data)
                    case _:
                        data = UNiiRawData(data)
            except ValueError as ex:
                data = None
                if (
                    self.command
                    in [
                        UNiiCommand.RESPONSE_REQUEST_INPUT_ARRANGEMENT,
                        UNiiCommand.RESPONSE_REQUEST_OUTPUT_ARRANGEMENT,
                    ]
                    and str(ex) == "Invalid block number"
                ):
                    pass
                else:
                    logger.error(ex)
                    logger.debug("Payload: 0x%s", payload.hex())
                    raise UNiiInvallidMessageError() from ex
            except LookupError as ex:
                logger.error(ex)
                logger.debug("Payload: 0x%s", payload.hex())
                data = None
                raise UNiiInvallidMessageError() from ex
            # Catch all exceptions while in development, to be removed once stable.
            # pylint: disable=broad-exception-caught
            except Exception as ex:
                logger.error(ex)
                logger.debug("Payload: 0x%s", payload.hex())
                data = None
                raise UNiiInvallidMessageError() from ex
        self.data = data

    def _decrypt(self, shared_key: bytes, initial_value: bytes, payload: bytes):
        aes = AES.new(shared_key, AES.MODE_CTR, initial_value=initial_value, nonce=b"")

        return aes.decrypt(payload)
